import { Annotated, Bytes, Set, Dictionary, Fold, fold, Record, Tuple, Value, stringify } from "@preserves/core";
import { brackets, Item, parens, seq } from "./block";

export function sourceCodeFor(v: Value<never>): Item {
    return fold(v, {
        boolean(b: boolean): Item { return b.toString(); },
        single(f: number): Item { return f.toString(); },
        double(f: number): Item { return f.toString(); },
        integer(i: number): Item { return i.toString(); },
        string(s: string): Item { return JSON.stringify(s); },
        bytes(b: Bytes): Item {
            return seq(`Uint8Array.from(`, brackets(... Array.from(b).map(b => b.toString())), `)`);
        },
        symbol(s: symbol): Item { return `Symbol.for(${JSON.stringify(s.description!)})`; },

        record(r: Record<Value<never>, Tuple<Value<never>>, never>, k: Fold<never, Item>): Item {
            return seq(`_.Record<_val, _.Tuple<_val>, _ptr>`, parens(k(r.label), brackets(... r.map(k))));
        },
        array(a: Array<Value<never>>, k: Fold<never, Item>): Item {
            return brackets(... a.map(k));
        },
        set(s: Set<never>, k: Fold<never, Item>): Item {
            return seq('new _.Set<_val>', parens(brackets(... Array.from(s).map(k))));
        },
        dictionary(d: Dictionary<never>, k: Fold<never, Item>): Item {
            return seq('new _.Dictionary<_ptr>', parens(brackets(... Array.from(d).map(([kk,vv]) =>
                brackets(k(kk), k(vv))))));
        },

        annotated(a: Annotated<never>, k: Fold<never, Item>): Item {
            return seq('_.annotate<_ptr>', parens(k(a.item), ... a.annotations.map(k)));
        },

        pointer(t: never, _k: Fold<never, Item>): Item {
            throw new Error(`Cannot emit source code for construction of pointer ${stringify(t)}`);
        },
    });
}
